/**************************************************************************
Etherboot -  BOOTP/TFTP Bootstrap Program
3c507 NIC driver for Etherboot
***************************************************************************/

/* to get some global routines like printf */
#include "etherboot.h"
/* to get the interface to the body of the program */
#include "nic.h"

#define	DEBUG

/* Sources of information:

   Donald Becker's excellent 3c507 driver in Linux
   Intel 82596 data sheet (yes, 82596; it has a 586 compatibility mode)
*/

/* Code below stolen wholesale from 3c507.c driver in Linux */

/*
  			Details of the i82586.

   You'll really need the databook to understand the details of this part,
   but the outline is that the i82586 has two separate processing units.
   Both are started from a list of three configuration tables, of which only
   the last, the System Control Block (SCB), is used after reset-time.  The SCB
   has the following fields:
		Status word
		Command word
		Tx/Command block addr.
		Rx block addr.
   The command word accepts the following controls for the Tx and Rx units:
  */

#define	 CUC_START	 0x0100
#define	 CUC_RESUME	 0x0200
#define	 CUC_SUSPEND 0x0300
#define	 RX_START	 0x0010
#define	 RX_RESUME	 0x0020
#define	 RX_SUSPEND	 0x0030

/* The Rx unit uses a list of frame descriptors and a list of data buffer
   descriptors.  We use full-sized (1518 byte) data buffers, so there is
   a one-to-one pairing of frame descriptors to buffer descriptors.

   The Tx ("command") unit executes a list of commands that look like:
		Status word		Written by the 82586 when the command is done.
		Command word	Command in lower 3 bits, post-command action in upper 3
		Link word		The address of the next command.
		Parameters		(as needed).

	Some definitions related to the Command Word are:
 */
#define CMD_EOL		0x8000			/* The last command of the list, stop. */
#define CMD_SUSP	0x4000			/* Suspend after doing cmd. */
#define CMD_INTR	0x2000			/* Interrupt after doing cmd. */

enum commands {
	CmdNOp = 0, CmdSASetup = 1, CmdConfigure = 2, CmdMulticastList = 3,
	CmdTx = 4, CmdTDR = 5, CmdDump = 6, CmdDiagnose = 7};

/* Information that need to be kept for each board. */
struct net_local {
	int last_restart;
	unsigned short rx_head;
	unsigned short rx_tail;
	unsigned short tx_head;
	unsigned short tx_cmd_link;
	unsigned short tx_reap;
};

/*
  		Details of the EtherLink16 Implementation
  The 3c507 is a generic shared-memory i82586 implementation.
  The host can map 16K, 32K, 48K, or 64K of the 64K memory into
  0x0[CD][08]0000, or all 64K into 0xF[02468]0000.
  */

/* Offsets from the base I/O address. */
#define	SA_DATA		0	/* Station address data, or 3Com signature. */
#define MISC_CTRL	6	/* Switch the SA_DATA banks, and bus config bits. */
#define RESET_IRQ	10	/* Reset the latched IRQ line. */
#define SIGNAL_CA	11	/* Frob the 82586 Channel Attention line. */
#define ROM_CONFIG	13
#define MEM_CONFIG	14
#define IRQ_CONFIG	15
#define EL16_IO_EXTENT 16

/* The ID port is used at boot-time to locate the ethercard. */
#define ID_PORT		0x100

/* Offsets to registers in the mailbox (SCB). */
#define iSCB_STATUS	0x8
#define iSCB_CMD		0xA
#define iSCB_CBL		0xC	/* Command BLock offset. */
#define iSCB_RFA		0xE	/* Rx Frame Area offset. */

/*  Since the 3c507 maps the shared memory window so that the last byte is
	at 82586 address FFFF, the first byte is at 82586 address 0, 16K, 32K, or
	48K corresponding to window sizes of 64K, 48K, 32K and 16K respectively. 
	We can account for this be setting the 'SBC Base' entry in the ISCP table
	below for all the 16 bit offset addresses, and also adding the 'SCB Base'
	value to all 24 bit physical addresses (in the SCP table and the TX and RX
	Buffer Descriptors).
					-Mark	
	*/
 
/*
  What follows in 'init_words[]' is the "program" that is downloaded to the
  82586 memory.	 It's mostly tables and command blocks, and starts at the
  reset address 0xfffff6.  This is designed to be similar to the EtherExpress,
  thus the unusual location of the SCB at 0x0008.

  Even with the additional "don't care" values, doing it this way takes less
  program space than initializing the individual tables, and I feel it's much
  cleaner.

  The databook is particularly useless for the first two structures, I had
  to use the Crynwr driver as an example.

   The memory setup is as follows:
   */

#define CONFIG_CMD	0x0018
#define SET_SA_CMD	0x0024
#define SA_OFFSET	0x002A
#define IDLELOOP	0x30
#define TDR_CMD		0x38
#define TDR_TIME	0x3C
#define DUMP_CMD	0x40
#define DIAG_CMD	0x48
#define SET_MC_CMD	0x4E
#define DUMP_DATA	0x56	/* A 170 byte buffer for dump and Set-MC into. */

#define TX_BUF_START	0x0100
#define TX_BUF_SIZE 	(1518+14+20+16) /* packet+header+TBD */

#define RX_BUF_START	0x0800
#define RX_BUF_SIZE 	(1518+14+18)	/* packet+header+RBD */
#define RX_BUF_END	(mem_end - mem_start)

/*
  That's it: only 86 bytes to set up the beast, including every extra
  command available.  The 170 byte buffer at DUMP_DATA is shared between the
  Dump command (called only by the diagnostic program) and the SetMulticastList
  command. 

  To complete the memory setup you only have to write the station address at
  SA_OFFSET and create the Tx & Rx buffer lists.

  The Tx command chain and buffer list is setup as follows:
  A Tx command table, with the data buffer pointing to...
  A Tx data buffer descriptor.  The packet is in a single buffer, rather than
	chaining together several smaller buffers.
  A NoOp command, which initially points to itself,
  And the packet data.

  A transmit is done by filling in the Tx command table and data buffer,
  re-writing the NoOp command, and finally changing the offset of the last
  command to point to the current Tx command.  When the Tx command is finished,
  it jumps to the NoOp, when it loops until the next Tx command changes the
  "link offset" in the NoOp.  This way the 82586 never has to go through the
  slow restart sequence.

  The Rx buffer list is set up in the obvious ring structure.  We have enough
  memory (and low enough interrupt latency) that we can avoid the complicated
  Rx buffer linked lists by alway associating a full-size Rx data buffer with
  each Rx data frame.

  I currently use one transmit buffer starting at TX_BUF_START (0x0100), and
  use the rest of memory, from RX_BUF_START to RX_BUF_END, for Rx buffers.

  */

static unsigned short init_words[] = {
	/*	System Configuration Pointer (SCP). */
	0x0000,					/* Set bus size to 16 bits. */
	0,0,					/* pad words. */
	0x0000,0x0000,			/* ISCP phys addr, set in init_82586_mem(). */

	/*	Intermediate System Configuration Pointer (ISCP). */
	0x0001,					/* Status word that's cleared when init is done. */
	0x0008,0,0,				/* SCB offset, (skip, skip) */

	/* System Control Block (SCB). */
	0,0xf000|RX_START|CUC_START,	/* SCB status and cmd. */
	CONFIG_CMD,				/* Command list pointer, points to Configure. */
	RX_BUF_START,				/* Rx block list. */
	0,0,0,0,				/* Error count: CRC, align, buffer, overrun. */

	/* 0x0018: Configure command.  Change to put MAC data with packet. */
	0, CmdConfigure,		/* Status, command.		*/
	SET_SA_CMD,				/* Next command is Set Station Addr. */
	0x0804,					/* "4" bytes of config data, 8 byte FIFO. */
	0x2e40,					/* Magic values, including MAC data location. */
	0,						/* Unused pad word. */

	/* 0x0024: Setup station address command. */
	0, CmdSASetup,
	SET_MC_CMD,				/* Next command. */
	0xaa00,0xb000,0x0bad,	/* Station address (to be filled in) */

	/* 0x0030: NOP, looping back to itself.	 Point to first Tx buffer to Tx. */
	0, CmdNOp, IDLELOOP, 0 /* pad */,

	/* 0x0038: A unused Time-Domain Reflectometer command. */
	0, CmdTDR, IDLELOOP, 0,

	/* 0x0040: An unused Dump State command. */
	0, CmdDump, IDLELOOP, DUMP_DATA,

	/* 0x0048: An unused Diagnose command. */
	0, CmdDiagnose, IDLELOOP,

	/* 0x004E: An empty set-multicast-list command. */
	0, CmdMulticastList, IDLELOOP, 0,
};

/* NIC specific static variables go here */

static unsigned short		ioaddr, irq, scb_base;
static Address			mem_start, mem_end;
static char			if_port;

#define	read_mem(m,s)	fbcopy(m, (char *)s, sizeof(s))

static void setup_rx_buffer(struct nic *nic)
{
	static unsigned short	rx_cmd[16] = {
		0x0000,			/* Rx status */
		0xC000,			/* Rx command, only and last */
		RX_BUF_START,		/* Link back to itself */
		RX_BUF_START + 22,	/* Buffer offset */
		0x0000, 0x0000, 0x0000,	/* Pad for dest addr */
		0x0000, 0x0000, 0x0000,	/* Pad for source addr */
		0x0000,			/* Pad for protocol */
		0x0000,			/* Buffer: Actual count */
		-1,			/* Buffer: Next (none) */
		RX_BUF_START + 0x20,	/* Buffer: Address low (+ scb_base) */
		0x0000,			/* Buffer: Address high */
		0x8000 + RX_BUF_SIZE - 0x20
	};

	/* adjust some contents */
	rx_cmd[13] = scb_base + RX_BUF_START + 0x20;
#ifdef	DEBUG
	{
	int	i;
	for (i = 0; i < 16; ++i)
		printf("%x ", rx_cmd[i]);
	printf("\r\n");
	}
#endif
#ifdef	ETHERBOOT32
	bcopy((char *)rx_cmd, (char *)mem_start + RX_BUF_START, sizeof(rx_cmd));
#endif
#ifdef	ETHERBOOT16
	bcopyf((char *)rx_cmd, (char *)mem_start + RX_BUF_START, sizeof(rx_cmd));
#endif
}

/**************************************************************************
RESET - Reset adapter
***************************************************************************/
static void t507_reset(struct nic *nic)
{
	int		boguscnt;
#ifdef	ETHERBOOT32
	unsigned short	*shmem = (short *)mem_start;
#endif
#ifdef	ETHERBOOT16
	unsigned short	shmem[CONFIG_CMD>>1];
#endif

	/* put the card in its initial state */

	/* Enable loopback to protect the wire while starting up,
	   and hold the 586 in reset during the memory initialisation. */
	OUTB(ioaddr + MISC_CTRL, 0x20);

	/* Fix the ISCP address and base. */
	init_words[3] = scb_base;
	init_words[7] = scb_base;

	/* Write the words at 0xfff6. */
	/* Write the words at 0x0000. */
	/* Fill in the station address. */
#ifdef	ETHERBOOT32
	bcopy((char *)init_words, (char *)(mem_end - 10), 10);
	bcopy((char *)&init_words[5], (char *)mem_start, sizeof(init_words) - 10);
	bcopy(nic->node_addr, (char *)mem_start + SA_OFFSET, ETHER_ADDR_SIZE);
#endif
#ifdef	ETHERBOOT16
	bcopyf((char *)init_words, mem_end - 10, 10);
	bcopyf((char *)&init_words[5], mem_start, sizeof(init_words) - 10);
	bcopyf(nic->node_addr, mem_start + SA_OFFSET, ETHER_ADDR_SIZE);
#endif
	setup_rx_buffer(nic);

	/* Start the 586 by releasing the reset line, but leave loopback. */
	OUTB(ioaddr + MISC_CTRL, 0xA0);

	/* This was time consuming to track down; you need to give two channel
	   attention signals to reliably start up the i82586. */
	OUTB(ioaddr + SIGNAL_CA, 0);
	boguscnt = 50;
	while (
#ifdef	ETHERBOOT16
		read_mem(mem_start, shmem),
#endif
		shmem[iSCB_STATUS>>1] == 0)
	{
		if (--boguscnt == 0)
		{
			printf("i82586 initialisation timed out with status %x, cmd %x\r\n",
				shmem[iSCB_STATUS>>1], shmem[iSCB_CMD>>1]);
			break;
		}
	}
	/* Issue channel-attn -- the 82586 won't start. */
	OUTB(ioaddr + SIGNAL_CA, 0);

	/* Disable loopback. */
	OUTB(ioaddr + MISC_CTRL, 0x80);
#ifdef	DEBUG
#ifdef	ETHERBOOT16
	read_mem(mem_start, shmem);
#endif
	printf("i82586 status %x, cmd %x\r\n",
		shmem[iSCB_STATUS>>1], shmem[iSCB_CMD>>1]);
#endif	DEBUG
}

/**************************************************************************
POLL - Wait for a frame
***************************************************************************/
static int t507_poll(struct nic *nic)
{
#ifdef	ETHERBOOT32
	unsigned short	*shmem = (short *)mem_start + RX_BUF_START;
#endif
#ifdef	ETHERBOOT16
	unsigned short	shmem[16];
#endif

	/* return true if there's an ethernet packet ready to read */
	if (
#ifdef	ETHERBOOT16
		read_mem(mem_start + RX_BUF_START, shmem),
#endif
		(shmem[0] & 0x2000) == 0)
		return (0);
	/* We have a frame, copy it to our buffer */
#ifdef	DEBUG
	printf("Size %x Count %d\r\n", shmem[15], shmem[11] & 0x3fff);
	{
	int	i;
	for (i = 0; i < 16; ++i)
		printf("%x ", shmem[i]);
	printf("\r\n");
	}
#endif
#ifdef	ETHERBOOT32
	bcopy((char *)mem_start + RX_BUF_START + 0x20, nic->packet, 1500);
#endif
#ifdef	ETHERBOOT16
	fbcopy(mem_start + RX_BUF_START + 0x20, nic->packet, 1500);
#endif
	return (1);
}

/**************************************************************************
TRANSMIT - Transmit a frame
***************************************************************************/
static void t507_transmit(
struct nic *nic,
char *d,			/* Destination */
unsigned int t,			/* Type */
unsigned int s,			/* size */
char *p)			/* Packet */
{
	Address			bptr;
	unsigned short		type, z;
	static unsigned short	tx_cmd[11] = {
		0x0,			/* Tx status */
		CmdTx,			/* Tx command */
		TX_BUF_START+16,	/* Next command is a NoOp */
		TX_BUF_START+8,		/* Data Buffer offset */
		0x8000,			/* | with size */
		0xffff,			/* No next data buffer */
		TX_BUF_START+22,	/* + scb_base */
		0x0,			/* Buffer address high bits (always zero) */
		0x0,			/* Nop status */
		CmdNOp,			/* Nop command */
		TX_BUF_START+16		/* Next is myself */
	};
#ifdef	ETHERBOOT32
	unsigned short	*shmem = (short *)mem_start + TX_BUF_START;
#endif
#ifdef	ETHERBOOT16
	unsigned short	shmem[11];
#endif

	/* send the packet to destination */
	/* adjust some contents */
	type = htons(t);
	tx_cmd[4] = (s + ETHER_HDR_SIZE) | 0x8000;
	tx_cmd[6] = TX_BUF_START + 22 + scb_base;
	bptr = mem_start + TX_BUF_START;
#ifdef	ETHERBOOT32
	bcopy((char *)tx_cmd, (char *)bptr, sizeof(tx_cmd));
	bptr += sizeof(tx_cmd);
	bcopy(d, (char *)bptr, ETHER_ADDR_SIZE);
	bptr += ETHER_ADDR_SIZE;
	bcopy(nic->node_addr, (char *)bptr, ETHER_ADDR_SIZE);
	bptr += ETHER_ADDR_SIZE;
	bcopy((char *)&type, (char *)bptr, sizeof(type));
	bptr += sizeof(type);
	bcopy(p, (char *)bptr, s);
	/* Change the offset in the IDLELOOP */
	*(unsigned short *)(mem_start + IDLELOOP + 4) = TX_BUF_START;
#endif
#ifdef	ETHERBOOT16
	bcopyf((char *)tx_cmd, bptr, sizeof(tx_cmd));
	bptr += sizeof(tx_cmd);
	bcopyf(d, bptr, ETHER_ADDR_SIZE);
	bptr += ETHER_ADDR_SIZE;
	bcopyf(nic->node_addr, bptr, ETHER_ADDR_SIZE);
	bptr += ETHER_ADDR_SIZE;
	bcopyf((char *)&type, bptr, sizeof(type));
	bptr += sizeof(type);
	bcopyf(p, bptr, s);
	/* Change the offset in the IDLELOOP */
	z = TX_BUF_START;
	bcopyf((char *)&z, mem_start + IDLELOOP + 4, sizeof(z));
#endif
	/* Wait for transmit completion */
	while (
#ifdef	ETHERBOOT16
		read_mem(mem_start + TX_BUF_START, shmem),
#endif
		(shmem[0] & 0x4000) == 0)
		;
	/* Change the offset in the IDLELOOP back and
	   change the final loop to point here */
#ifdef	ETHERBOOT32
	*(unsigned short *)(mem_start + IDLELOOP + 4) = IDLELOOP;
	*(unsigned short *)(mem_start + TX_BUF_START + 20) = IDLELOOP;
#endif
#ifdef	ETHERBOOT16
	z = IDLELOOP;
	bcopyf((char *)&z, mem_start + IDLELOOP + 4, sizeof(z));
	bcopyf((char *)&z, mem_start + TX_BUF_START + 20, sizeof(z));
#endif
#ifdef	DEBUG
	read_mem(mem_start + TX_BUF_START, shmem);
	{
	int	i;
	for (i = 0; i < 11; ++i)
		printf("%x ", shmem[i]);
	printf("\r\n");
	}
#endif
}

/**************************************************************************
DISABLE - Turn off ethernet interface
***************************************************************************/
static void t507_disable(struct nic *nic)
{
}

static int t507_probe1(struct nic *nic, unsigned short ioaddr)
{
	int			i;
	Address			size;
	char			mem_config;

	if (inb(ioaddr) != '*' || inb(ioaddr+1) != '3'
		|| inb(ioaddr+2) != 'C' || inb(ioaddr+3) != 'O')
		return (0);
	irq = inb(ioaddr + IRQ_CONFIG) & 0x0f;
	mem_config = inb(ioaddr + MEM_CONFIG);
	if (mem_config & 0x20)
	{
		size = 65536L;
		mem_start = 0xf00000L + (mem_config & 0x08 ? 0x080000L
			: (((Address)mem_config & 0x3) << 17));
	}
	else
	{
		size = ((((Address)mem_config & 0x3) + 1) << 14) & 0xffffL;
		mem_start = 0x0c0000L + (((Address)mem_config & 0x18) << 12);
	}
	mem_end = mem_start + size;
	scb_base = 65536L - size;
	if_port = inb(ioaddr + ROM_CONFIG) & 0x80;
	printf("\r\n3c507 ioaddr 0x%x, IRQ %d, mem [0x%X-0x%X], %sternal xcvr, addr ",
		ioaddr, irq, mem_start, mem_end, if_port ? "in" : "ex");
	/* Get station address */
	OUTB(ioaddr + MISC_CTRL, 0x01);
	for (i = 0; i < ETHER_ADDR_SIZE; ++i)
	{
		printf("%b", nic->node_addr[i] = inb(ioaddr+i));
		if (i < ETHER_ADDR_SIZE -1)
			printf(":");
	}
	printf("\r\n");
	return (1);
}

/**************************************************************************
PROBE - Look for an adapter, this routine's visible to the outside
***************************************************************************/
struct nic *t507_probe(struct nic *nic, unsigned short *probe_addrs)
{
	static unsigned char	init_ID_done = 0;
	unsigned short		lrs_state = 0xff;
	static unsigned short	io_addrs[] = { 0x300, 0x320, 0x340, 0x280, 0 };
	unsigned short		*p;
	int			i;

	if (init_ID_done == 0)
	{
		/* Send the ID sequence to the ID_PORT to enable the board */
		OUTB(ID_PORT, 0x00);
		for (i = 0; i < 255; ++i)
		{
			OUTB(ID_PORT, lrs_state);
			lrs_state <<= 1;
			if (lrs_state & 0x100)
				lrs_state ^= 0xe7;
		}
		OUTB(ID_PORT, 0x00);
		init_ID_done = 1;
	}
	/* if probe_addrs is 0, then routine can use a hardwired default */
	if (probe_addrs == 0)
		probe_addrs = io_addrs;
	for (p = probe_addrs; (ioaddr = *p) != 0; ++p)
		if (t507_probe1(nic, ioaddr))
			break;
	if (ioaddr != 0)
	{
		/* point to NIC specific routines */
		t507_reset(nic);
		nic->reset = t507_reset;
		nic->poll = t507_poll;
		nic->transmit = t507_transmit;
		nic->disable = t507_disable;
		return nic;
	}
	/* else */
	{
		return 0;
	}
}
